"""
Test instance creation.
"""

import os

from shutil import rmtree

import tiddlywebwiki.instancer as instancer

from tiddlyweb.config import config as system_config


instance_dir = 'test_instance'
config = {
    'instance_tiddlers': {}
}


def setup_module(module):
    try:
        rmtree(instance_dir)
    except:
        pass


class TestCreateInstance(object):

    def teardown_method(self, module):
        os.chdir('..')
        rmtree(instance_dir)

    def test_create_default(self):
        instancer.create_instance(instance_dir, config, system_config=system_config)
        contents = _get_file_contents('../%s/tiddlywebconfig.py' % instance_dir)

        assert 'config = {\n' in contents
        assert "'secret': '" in contents
        assert '\n}' in contents

    def test_create_custom(self):
        defaults = {
            'foo': 'lorem',
            'bar': 'ipsum'
        }
        instancer.create_instance(instance_dir, config, defaults=defaults,
                system_config=system_config)
        contents = _get_file_contents('../%s/tiddlywebconfig.py' % instance_dir)

        assert 'config = {\n' in contents
        assert "'foo': 'lorem'" in contents
        assert "'bar': 'ipsum'" in contents
        assert '\n}' in contents


def _get_file_contents(filepath):
    f = open(filepath)
    contents = f.read()
    f.close()
    return contents



"""
A built-in twanager plugin for setting up a new instance
of TiddlyWeb, to lighten the load on getting started.

Example:

    twanager instance foo

This will create a foo directory containing:

    * empty tiddlywebconfig.py
    * store directory with a system bag containing
      the important plugins (assuming text store is
      in default config.py)
"""

import os
import random
import time

from tiddlyweb.config import config
from tiddlyweb.util import sha
from tiddlyweb.model.bag import Bag
from tiddlyweb.model.recipe import Recipe
from tiddlywebwiki.importer import import_list
from tiddlywebplugins.utils import get_store


CONFIG_NAME = 'tiddlywebconfig.py'


def instance(directory, system_config=None):
    """Create a TiddlyWebWiki instance in the given directory"""
    if not directory:
        raise ValueError('You must provide the name of a directory.')
    if os.path.exists(directory):
        raise IOError('Your chosen directory already exists. Choose a different name.')

    if system_config == None:
        system_config = {}
    store = get_store(system_config)

    cfg = {
        'system_plugins': ['tiddlywebwiki', 'tiddlywebplugins.status', 'differ'],
        'twanager_plugins': ['tiddlywebwiki']
    }
    create_instance(directory, config, defaults=cfg, system_config=system_config)

    bag = Bag('system')
    bag.policy.write = ['R:ADMIN']
    bag.policy.create = ['R:ADMIN']
    bag.policy.delete = ['R:ADMIN']
    bag.policy.manage = ['R:ADMIN']
    bag.policy.accept = ['R:ADMIN']
    store.put(bag)

    bag = Bag('common')
    bag.policy.delete = ['R:ADMIN']
    bag.policy.manage = ['R:ADMIN']
    store.put(bag)

    recipe = _make_recipe('default', ['system', 'common'], store)


def create_instance(directory, cfg, defaults=None, system_config=None):
    """
    Create a TiddlyWeb instance directory.

    accepts an optional dictionary with default configuration values
    """
    os.mkdir(directory)
    os.chdir(directory)
    _generate_config(defaults)
    store = get_store(system_config)
    instance_tiddlers = cfg.get('instance_tiddlers', {})
    bag_names = [bag for bag, tiddlers in instance_tiddlers.items()]
    for bag in bag_names:
        bag = Bag(bag)
        store.put(bag)
    update_instance(system_config)


def update_instance(cfg=None):
    """
    Update a TiddlyWeb instance by reimporting instance_tiddlers.
    """
    if cfg == None:
        cfg = {}
    instance_tiddlers = cfg.get('instance_tiddlers', {})
    [import_list(bag, tiddlers, cfg) for bag, tiddlers
        in instance_tiddlers.items()]


def _generate_config(defaults=None):
    """
    Write a default tiddlywebconfig.py to the CWD.

    accepts an optional dictionary with configuration values
    """
    intro = '%s\n%s' % ('# A basic configuration.',
        "# Run 'pydoc tiddlyweb.config' for details on configuration items.")
    cfg = {
        'secret': _generate_secret()
    }
    cfg.update(defaults or {})

    config_string = 'config = %s\n' % _pretty_print(cfg)
    cfg = open(CONFIG_NAME, 'w')
    cfg.write('%s\n%s' % (intro, config_string))
    cfg.close()


def _pretty_print(dic): # TODO: use pprint?
    """
    generate an indented string representation of a dictionary
    """
    def escape_strings(value):
        if hasattr(value, "join"): # XXX: checking for join method hacky!?
            return "'%s'" % value
        else:
            return value
    lines = ["    '%s': %s" % (k, escape_strings(v)) for k, v in dic.items()] # TODO: use double quotes (for consistency with JSON in policy files)
    return '{\n%s\n}' % ',\n'.join(lines)


def _generate_secret():
    """
    Create a pseudo-random secret to be used for message authentication.
    """
    digest = sha(str(time.time()))
    digest.update(str(random.random()))
    digest.update('tiddlyweb and tiddlywiki are rad')
    return digest.hexdigest()


def _make_recipe(recipe_name, bags, store):
    """Make a recipe with recipe_name."""
    recipe = Recipe(recipe_name)
    recipe_list = [[bag, ''] for bag in bags]
    recipe.set_recipe(recipe_list)
    store.put(recipe)



"""
A TiddlyWeb plugin providing a multi-user TiddlyWiki environment.
"""

__version__ = '0.15.0'


def init(config):
    # These imports must be in init otherwise getting
    # the version number causes cyclic imports to happen.
    import tiddlywebwiki.manage
    import tiddlywebplugins.atom
    from tiddlyweb.config import merge_config
    from tiddlywebwiki.config import config as twwconfig
    merge_config(config, twwconfig)
    tiddlywebwiki.manage.init(config)
    tiddlywebplugins.atom.init(config)
    # XXX and add selector for POST a wiki



"""
Functionality for parsing TiddlyWiki documents.
"""

import codecs

import html5lib
from html5lib import treebuilders

from tiddlyweb.model.tiddler import Tiddler, string_to_tags_list


def import_wiki(environ, start_response):
    """
    Accept a tiddlywiki as POST and using it as the source
    parse it for tiddlers to be stored in the named bag.
    """
    bag_name = _determine_bag_name(environ)
    bag = _get_bag(environ, bag_name, True)
    length = environ['CONTENT_LENGTH']
    content = environ['wsgi.input'].read(int(length))

    bag.policy.allows(environ['tiddlyweb.usersign'], 'create')

    try:
        serialize_type, mime_type = web.get_serialize_type(environ)
        serializer = Serializer(serialize_type, environ)
        serializer.object = bag

        serializer.from_string(content)
    except NoSerializationError:
        raise HTTP415('Content type not supported: %s' % mime_type)
    except AttributeError, exc:
        raise HTTP400('Content malformed: %s' % exc)

    start_response("204 No Content",
            [('Location', '%s/tiddlers' % web.bag_url(environ, bag))])
    return ['']


def import_wiki_file(store, filename='wiki', bagname='wiki'):
    """
    Read a wiki in a file and import all the tiddlers into a bag.
    """
    wikifile = codecs.open(filename, encoding='utf-8', errors='replace')
    wikitext = wikifile.read()
    wikifile.close()
    return import_wiki(store, wikitext, bagname)


def import_wiki(store, wikitext, bagname='wiki'):
    """
    Import the wiki provided as a string and import all the tiddlers
    into a bag.
    """
    parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder('beautifulsoup'))
    soup = parser.parse(wikitext)
    store_area = soup.find('div', id='storeArea')
    divs = store_area.findAll('div')

    for tiddler_div in divs:
        handle_tiddler_div(bagname, tiddler_div, store)


def handle_tiddler_div(bagname, tiddler_div, store):
    """
    Create a new Tiddler from a tiddler div, in BeautifulSoup form.
    """
    tiddler = get_tiddler_from_div(tiddler_div)
    tiddler.bag = bagname
    try:
        store.put(tiddler)
    except OSError, exc:
        # This tiddler has a name that we can't yet write to the
        # store. For now we just state the error and carry on.
        import sys
        print >> sys.stderr, 'Unable to write %s: %s' % (tiddler.title, exc)


def get_tiddler_from_div(node):
    """
    Create a Tiddler from a BeautifulSoup DIV node
    """
    tiddler = Tiddler(node['title'])
    try:
        tiddler.text = _html_decode(node.find('pre').contents[0])
    except IndexError:
        # there are no contents in the tiddler
        tiddler.text = ''

    for attr, value in node.attrs:
        data = node.get(attr, None)
        if data and attr != 'tags':
            if attr in (['modifier', 'created', 'modified']):
                tiddler.__setattr__(attr, data)
            elif (attr not in ['title', 'changecount'] and
                not attr.startswith('server.')):
                tiddler.fields[attr] = data
    if not node.get('modified', None) and tiddler.created:
        tiddler.modified = tiddler.created
    tiddler.tags = _tag_string_to_list(node.get('tags', ''))

    return tiddler


def _tag_string_to_list(string):
    """
    Turn a string of tags in TiddlyWiki format into a list.
    """
    return string_to_tags_list(string)


def _html_decode(text):
    """
    Decode HTML entities used in TiddlyWiki content into the 'real' things.
    """
    return text.replace('&gt;', '>').replace('&lt;', '<').replace('&amp;', '&').replace('&quot;', '"')



"""
Tests for turning a Tiddler into HTML within a TiddlyWiki document
"""

from tiddlyweb.model.tiddler import Tiddler
from tiddlyweb.serializer import Serializer
from tiddlyweb.config import config


# TODO: use module_setup!?
BASE_TIDDLYWIKI = 'tiddlywebwiki/empty.html'
config['base_tiddlywiki'] = BASE_TIDDLYWIKI
environ = { 'tiddlyweb.config': config }


def test_html_attribute_escape():
    tiddler = Tiddler('escape "double" quotes in tiddler field values')
    tiddler.bag = 'foo "bar" baz'
    tiddler.recipe = 'baz "bar" foo'
    tiddler.modifier = 'Chris "sensei" Dent'
    tiddler.tags = ["foo", 'xxx "yyy" zzz']
    tiddler.fields["custom"] = u"""lorem 'ipsum' dolor "sit" amet"""
    tiddler.text = ''
    serializer = Serializer('tiddlywebwiki.serialization', environ)
    serializer.object = tiddler
    string = serializer.to_string()

    assert r'''title="escape &quot;double&quot; quotes in tiddler field values"''' in string
    assert r'''server.title="escape &quot;double&quot; quotes in tiddler field values"''' in string
    assert r'''bag="foo &quot;bar&quot; baz"''' in string
    assert r'''recipe="baz &quot;bar&quot; foo"''' in string
    assert r'''server.workspace="bags/foo &quot;bar&quot; baz"''' in string
    assert r'''modifier="Chris &quot;sensei&quot; Dent"''' in string
    assert r'''creator="Chris &quot;sensei&quot; Dent"''' in string
    assert r'''tags="foo [[xxx &quot;yyy&quot; zzz]]"''' in string
    assert r'''custom="lorem 'ipsum' dolor &quot;sit&quot; amet"''' in string



"""
Test reading in a tiddler div from a TiddlyWiki document.
"""

import re

import html5lib

from html5lib import treebuilders

from tiddlyweb.config import config
from tiddlyweb.model.bag import Bag
from tiddlyweb.model.tiddler import Tiddler
from tiddlyweb.store import Store

from tiddlywebwiki.tiddlywiki import handle_tiddler_div, get_tiddler_from_div
from tiddlywebwiki.importer import process_tiddler, handle_recipe


BAGNAME = 'test'
SAMPLE_BASIC_TIDDLER = """
<div title="GettingStarted">
<pre>To get started with this blank [[TiddlyWiki]], you'll need to modify the following tiddlers:
* [[SiteTitle ]] & [[SiteSubtitle]]: The title and subtitle of the site, as shown above (after saving, they will also appear in the browser title bar)
* [[MainMenu]]: The menu (usually on the left)
* [[DefaultTiddlers]]: Contains the names of the tiddlers that you want to appear when the TiddlyWiki is opened You'll also need to enter your username for signing your edits: <<option txtUserName>></pre>
</div>
"""

SAMPLE_EMPTY_TIDDLER = """
<div title="GettingStopped">
<pre>
</pre>
</div>
"""


def setup_module(module):
    module.store = Store(config['server_store'][0], environ={'tiddlyweb.config': config})
    bag = Bag(BAGNAME)
    module.store.put(bag)


def test_import_simple_tiddler_div():
    div = _parse(SAMPLE_BASIC_TIDDLER)
    assert div['title'] == 'GettingStarted'

    handle_tiddler_div(BAGNAME, div, store)

    tiddler = Tiddler('GettingStarted', BAGNAME)
    tiddler = store.get(tiddler)
    assert tiddler.title == 'GettingStarted'
    assert 'as shown above (after' in tiddler.text

def test_import_empty_tiddler_div():
    div = _parse(SAMPLE_EMPTY_TIDDLER)
    assert div['title'] == 'GettingStopped'

    handle_tiddler_div(BAGNAME, div, store)

    tiddler = Tiddler('GettingStopped', BAGNAME)
    tiddler = store.get(tiddler)
    assert tiddler.title == 'GettingStopped'
    assert tiddler.text == ''

def test_handle_timestamps():
    tiddler_element = """
    <div title="Foo" created="200901011200" modified="200911261330">
    <pre></pre>
    </div>
    """
    div = _parse(tiddler_element)
    tiddler = get_tiddler_from_div(div)
    assert tiddler.created == '200901011200'
    assert tiddler.modified == '200911261330'

    tiddler_element = """
    <div title="Foo" created="200901011200">
    <pre></pre>
    </div>
    """
    div = _parse(tiddler_element)
    tiddler = get_tiddler_from_div(div)
    assert tiddler.created == '200901011200'
    assert tiddler.modified == '200901011200'

    tiddler_element = """
    <div title="Foo" modified="200911261330">
    <pre></pre>
    </div>
    """
    div = _parse(tiddler_element)
    tiddler = get_tiddler_from_div(div)
    assert tiddler.created == ''
    assert tiddler.modified == '200911261330'

    tiddler_element = """
    <div title="Foo">
    <pre></pre>
    </div>
    """
    div = _parse(tiddler_element)
    tiddler = get_tiddler_from_div(div)
    assert re.match('\d{12}', tiddler.modified)
    assert tiddler.created == ''


def test_omit_reserved_fields():
    tiddler_element = """
    <div title="Hello World" changecount="3" title="bar"
        server.host="example.org" server.workspace="default"
        custom="baz">
    <pre>lorem ipsum
    dolor sit amet</pre>
    </div>
    """
    div = _parse(tiddler_element)
    assert div['title'] == 'Hello World'

    handle_tiddler_div(BAGNAME, div, store)

    tiddler = Tiddler('Hello World', BAGNAME)
    tiddler = store.get(tiddler)
    assert tiddler.title == 'Hello World'
    assert tiddler.fields['custom'] == 'baz'
    assert tiddler.fields.has_key('title') == False
    assert tiddler.fields.has_key('server.host') == False
    assert tiddler.fields.has_key('server.workspace') == False
    assert tiddler.fields.has_key('changecount') == False

def test_importer_preprocessing():
    div = process_tiddler(SAMPLE_BASIC_TIDDLER)
    assert div['title'] == 'GettingStarted'
    assert '<<option txtUserName>>' in '%s' % div


def test_handle_recipe_basic():
    """
    handle_recipe takes a base url and some utf-8 encoded
    content that is a recipe. It returns a list of urls.
    """
    urls = handle_recipe('http://example.com/', """
tiddler: monkey/pirate.tiddler
recipe: special/thing.recipe
""")
    assert urls == ['http://example.com/monkey/pirate.tiddler', 'http://example.com/special/thing.recipe']


def test_handle_recipe_urls():
    """
    handle_recipe takes a base url and some utf-8 encoded
    content that is a recipe. It returns a list of urls.
    """
    urls = handle_recipe('http://example.com/', """
tiddler: %7Emonkey/pirate.tiddler
recipe: %7Especial/thing.recipe
""")
    assert urls == ['http://example.com/%7Emonkey/pirate.tiddler', 'http://example.com/%7Especial/thing.recipe']


def test_handle_recipe_utf8():
    """
    handle_recipe takes a base url and some utf-8 encoded
    content that is a recipe. It returns a list of urls.
    """
    urls = handle_recipe('http://example.com/', """
tiddler: monkey/%s/pirate.tiddler
recipe: special/%s/thing.recipe
""" % ('\xE2\x84\xA2', '\xE2\x84\xA2'))
    assert urls == ['http://example.com/monkey/%E2%84%A2/pirate.tiddler', 'http://example.com/special/%E2%84%A2/thing.recipe']


def test_handle_recipe_from_unicode():
    recipe = u"""
tiddler: monkey/%s/pirate.tiddler
recipe: special/%s/thing.recipe
""" % (u'\u2122', u'\u2122')
    recipe = recipe.encode('utf-8')
    urls = handle_recipe('http://example.com/', recipe)
    assert urls == ['http://example.com/monkey/%E2%84%A2/pirate.tiddler', 'http://example.com/special/%E2%84%A2/thing.recipe']


def test_handle_recipe_plugin():
    recipe = u'plugin: monkey/pirate.tid'
    recipe = recipe.encode('utf-8')
    urls = handle_recipe('http://example.com/', recipe)
    assert urls == ['http://example.com/monkey/pirate.tid']


def _parse(content):
    parser = html5lib.liberalxmlparser.XMLParser(tree=treebuilders.getTreeBuilder('beautifulsoup'))
    soup = parser.parseFragment(content)
    tiddler_div = soup.find('div')
    return tiddler_div



"""
A built-in twanager plugin for retrieving tiddlers, plugins or full
recipes from the TiddlyWiki Subversion repository. Provide the name of
one bag and one or more URIs (HTTP or file protocol) on the twanager
command line.

Example:

   twanager twimport myBag http://svn.tiddlywiki.org/Trunk/verticals/stunplugged/index.html.recipe

If the URL is a recipe it will be parsed for lines beginning with
"recipe:", "tiddler:" or "plugin:".

Recipes are retrieved and parsed recursively.

For tiddlers, if the end of the URI is ".js", then get the .js and .js.meta
files, massage them, join them together, make a tiddler, and put it in the
store.

Otherwise assume we have a tiddler in the <div> format and import it.
"""

import sys

import urllib
import urllib2
from urllib2 import urlopen, HTTPError
from urlparse import urljoin

import html5lib
from html5lib import treebuilders

from tiddlyweb.model.tiddler import Tiddler
from tiddlyweb.store import Store
from tiddlyweb.serializer import Serializer
from tiddlyweb.util import std_error_message
from tiddlywebwiki.tiddlywiki import handle_tiddler_div

from tiddlywebplugins.utils import get_store


def new_url2pathname(pathname):
    return pathname

urllib2.url2pathname = new_url2pathname


def import_list(bag, urls, config):
    """Import a list of URIs into bag."""
    for url in urls:
        import_one(bag, url, store=get_store(config))


def import_one(bag, url, store):
    """Import one URI into bag."""
    std_error_message("handling %s" % url)
    if url.endswith('.recipe'):
        import_via_recipe(bag, url, store)
    elif url.endswith('.js'):
        import_plugin(bag, url, store)
    elif url.endswith('.tid'):
        import_tid_tiddler(bag, url, store)
    else:
        import_tiddler(bag, url, store)


def import_via_recipe(bag, url, store):
    """
    Import one recipe into bag, calling import_one as needed.
    Will recurse recipes as it finds them. NO LOOP DETECTION.
    """
    recipe = get_url(url)
    recipe = recipe.encode('utf-8')
    urls = handle_recipe(url, recipe)
    for url in urls:
        import_one(bag, url, store)


def handle_recipe(url, content):
    """
    Take a url base and a UTF-8 encoded string and parse it
    as a TiddlyWiki cook recipe.
    """
    rules = [line for line in content.split('\n') if
            line.startswith('tiddler:') or
            line.startswith('plugin:') or
            line.startswith('recipe:')]
    urls = []
    for rule in rules:
        target = rule.split(':', 1)[1]
        target = target.lstrip().rstrip()
        if not (target.startswith('http:') or 
                target.startswith('https:')):
            if not '%' in target:
                target = urllib.quote(target)
            target_url = urljoin(url, target)
        else:
            target_url = target
        urls.append(target_url)
    return urls


def get_url(url):
    """
    Get the content at url, raising HTTPProblem if there is one.
    """
    try:
        std_error_message('getting url: %s' % url)
        return _get_url(url)
    except HTTPError, exc:
        std_error_message("HTTP Error while getting %s: %s" % (url, exc))
        sys.exit(1)


def _get_url(url):
    """
    Get the content at url, like get_url but without the
    exception handling.
    """
    getter = urlopen(url)
    content = getter.read().replace('\r', '')
    return unicode(content, 'utf-8')


def import_tid_tiddler(bag, url, store):
    """
    Import one tiddler, in the tid format, into bag.
    """
    content = get_url(url)
    tiddler_title = urllib.unquote(url.split('/')[-1])
    tiddler_title = _strip_extension(tiddler_title, '.tid')
    if not type(tiddler_title) == unicode:
        tiddler_title = unicode(tiddler_title, 'utf-8')
    tiddler = Tiddler(tiddler_title, bag)
    tiddler = process_tid_tiddler(tiddler, content)
    store.put(tiddler)


def process_tid_tiddler(tiddler, content):
    """
    Deserialize a tid.
    """
    serializer = Serializer('text')
    serializer.object = tiddler
    serializer.from_string(content)
    return tiddler


def import_tiddler(bag, url, store):
    """
    Import one tiddler into bag.
    """
    content = get_url(url)
    tiddler = process_tiddler(content)
    handle_tiddler_div(bag, tiddler, store)


def process_tiddler(content):
    """
    Turn some content into a div element representing
    a tiddler.
    """
    content = _escape_brackets(content)
    parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder('beautifulsoup'))
    soup = parser.parse(content)
    tiddler = soup.find('div')
    return tiddler


def import_plugin(bag, url, store):
    """
    Import one plugin into bag, retrieving both the
    .js and .js.meta files.

    If there is no meta file, then set title and tags
    to something appropriate before de-serializing.
    """
    meta_url = '%s.meta' % url
    plugin_content = get_url(url)
    default_title = _strip_extension(url, '.js').split('/')[-1]
    try:
        meta_content = _get_url(meta_url)
    except HTTPError:
        meta_content = 'title: %s\ntags: systemConfig\n' % default_title

    try:
        title = [line for line in meta_content.split('\n') if line.startswith('title:')][0]
        title = title.split(':', 1)[1].lstrip().rstrip()
    except IndexError:
        title = default_title
    tiddler_meta = '\n'.join([line for line in meta_content.split('\n') if not line.startswith('title:')])

    tiddler_meta.rstrip()
    tiddler_text = '%s\n\n%s' % (tiddler_meta, plugin_content)

    tiddler = Tiddler(title, bag)
    serializer = Serializer('text')
    serializer.object = tiddler
    serializer.from_string(tiddler_text)

    store.put(tiddler)


def _escape_brackets(content):
    open_pre = content.index('<pre>')
    close_pre = content.rindex('</pre>')
    start = content[0:open_pre+5]
    middle = content[open_pre+5:close_pre]
    end = content[close_pre:]
    middle = middle.replace('>', '&gt;').replace('<', '&lt;')
    return start + middle + end


def _strip_extension(name, ext):
    """
    Remove trailing extension from name.
    """
    ext_len = len(ext)
    if name[-ext_len:] == ext:
        name = name[:-ext_len]
    return name



"""
Test retrieving content from the TiddlyWiki Subversion repository.
"""

from tiddlywebwiki.importer import _strip_extension


def test_strip_extension():

    doc = """returns name without trailing extension"""
    actual = _strip_extension('foo.bar', '.bar')
    expected = 'foo'
    assert actual == expected, doc

    doc = """strips trailing extension only once"""
    actual = _strip_extension('foo.bar.bar', '.bar')
    expected = 'foo.bar'
    assert actual == expected, doc

    doc = """returns original name if extension does not appear at the end"""
    actual = _strip_extension('foo.bar.baz', '.bar')
    expected = 'foo.bar.baz'
    assert actual == expected, doc

    doc = """returns original name if extension is empty"""
    actual = _strip_extension('foo.bar', '')
    expected = 'foo.bar'
    assert actual == expected, doc

    doc = """raises TypeError for None extension"""
    try:
       _strip_extension('foo.bar', None)
       assert False, doc
    except TypeError:
        assert True, doc



import sys
import os

from tiddlyweb.model.bag import Bag
from tiddlyweb.store import Store
from tiddlyweb.manage import make_command, usage

from tiddlywebwiki.tiddlywiki import import_wiki_file
from tiddlywebwiki.instancer import update_instance
from tiddlywebwiki.importer import import_list


def init(config_in):
    global config
    config = config_in


@make_command()
def update(args):
    """Update all instance_tiddlers in the current instance."""
    update_instance(config)


@make_command()
def twimport(args):
    """Import one or more plugins, tiddlers or recipes in Cook format: <bag> <URI>"""
    bag = args[0]
    urls = args[1:]
    if not bag or not urls:
        raise IndexError('missing args')
    import_list(bag, urls, config)


@make_command()
def imwiki(args):
    """Import tiddlers from a Tiddlywiki document into a bag: <bag> <filename>"""
    store = _store()

    try:
        bag_name, filename = args[0:2]
        import_wiki_file(store, filename, bag_name)
    except IndexError, exc:
        usage("index error: %s" % exc)
    except ValueError, exc:
        usage("value error: %s" % exc)


def _store():
    """Get our Store from config."""
    return Store(config['server_store'][0],
            environ={'tiddlyweb.config': config})



"""
Differ
TiddlyWeb plugin to compare tiddler revisions

Usage:
  GET /diff?rev1=<tiddler>&rev2=<tiddler>[&format=<format>]
  POST /diff?rev1=<tiddler>[&format=<format>]

tiddler references are of the form bags/<title>[/<revision>]
(recipes are currently not supported in this context)

supported formats:
* human-readable line-by-line comparison (default; plain text)
* "unified" (plain text)
* "inline" (HTML)
* "horizontal" (side-by-side; HTML)

POST data (JSON representation of a tiddler) can be used instead of a tiddler
reference (rev1 or rev2 URL parameter)

TODO:
* unicode handling
* tests
"""

import cgi
import difflib

from tiddlyweb import control
from tiddlyweb.model.tiddler import Tiddler
from tiddlyweb.serializer import Serializer
from tiddlyweb.web import util as web
from tiddlyweb.web.http import HTTP400


__version__ = "0.4.0"


def init(config):
	config["selector"].add("/diff", GET=get_request, POST=post_request)


def get_request(environ, start_response):
	query = environ["tiddlyweb.query"]
	store = environ["tiddlyweb.store"]

	rev1_id = query.get("rev1", [None])[0]
	rev2_id = query.get("rev2", [None])[0]
	try:
		rev1 = _get_tiddler(rev1_id, store)
		rev2 = _get_tiddler(rev2_id, store)
	except AttributeError:
		raise HTTP400("missing revision parameter")

	format = query.get("format", [None])[0]

	content = compare_tiddlers(rev1, rev2, format)
	return _generate_response(content, environ, start_response)


def post_request(environ, start_response):
	length = int(environ["CONTENT_LENGTH"])
	post_content = environ["wsgi.input"].read(length)
	serializer = Serializer("json") # TODO: use content-type to determine serialization
	rev = Tiddler("untitled") # N.B.: Serializations need not contain title.
	serializer.object = rev
	serializer.from_string(post_content.decode("utf-8"))

	query = environ["tiddlyweb.query"]
	store = environ["tiddlyweb.store"]

	rev1_id = query.get("rev1", [None])[0]
	rev2_id = query.get("rev2", [None])[0]
	try:
		if not rev1_id:
			rev1 = rev
			rev2 = _get_tiddler(rev2_id, store)
		elif not rev2_id:
			rev1 = _get_tiddler(rev1_id, store)
			rev2 = rev
		else:
			raise HTTP400("ambiguous request")
	except AttributeError:
		raise HTTP400("missing revision parameter")

	format = query.get("format", [None])[0]

	content = compare_tiddlers(rev1, rev2, format)
	return _generate_response(content, environ, start_response)


def compare_tiddlers(rev1, rev2, format=None):
	"""
	compare two Tiddler instances
	"""
	serializer = Serializer("text")
	serializer.object = rev1
	rev1 = serializer.to_string()
	serializer.object = rev2
	rev2 = serializer.to_string()
	return diff(rev1, rev2, format)


def diff(a, b, format=None): # XXX: rename?
	"""
	create a diff representation of a string comparison

	defaults to human-readable line-by-line comparison
	alternative formats available are "unified", "horizontal" and "inline"
	"""
	if format == "unified":
		return "\n".join(difflib.unified_diff(a.splitlines(), b.splitlines()))
	elif format == "horizontal": # XXX: rename
		d = difflib.HtmlDiff()
		return d.make_file(a.splitlines(), b.splitlines()) # XXX: use make_table?!
	elif format == "inline":
		return generate_inline_diff(a, b)
	else:
		d = difflib.Differ() # XXX: use difflib.ndiff?
		result = list(d.compare(a.splitlines(), b.splitlines()))
		return "\n".join(result)


def generate_inline_diff(a, b): # TODO: optionally strip unchanged blocks
	"""
	compare two strings, highlighting differences inline

	returns a string using minimal HTML markup (INS and DEL elements)
	"""
	seq = difflib.SequenceMatcher(None, a, b)
	output = []
	for opcode, a0, a1, b0, b1 in seq.get_opcodes():
		if opcode == "equal":
			output.append(_html_transform(seq.a[a0:a1]))
		elif opcode == "insert":
			output.append("<ins>%s</ins>" % _html_transform(seq.b[b0:b1]))
		elif opcode == "delete":
			output.append("<del>%s</del>" % _html_transform(seq.a[a0:a1]))
		elif opcode == "replace":
			output.append("<del>%s</del><ins>%s</ins>" %
				(_html_transform(seq.a[a0:a1]), _html_transform(seq.b[b0:b1])))
		else:
			raise RuntimeError("unexpected opcode") # XXX: RuntimeError inappropriate?
	return "".join(output)


def _get_tiddler(id, store): # XXX: rename
	"""
	retrieve a tiddler revision based on an identifier

	returns the text serialization of the respective revision
	"""
	# retrieve tiddler revision from store
	type, name, title, rev = _resolve_identifier(id)
	if type == "bag":
		tiddler = Tiddler(title, name)
	else:
		raise HTTP400("recipes not supported") # TODO?
	tiddler.revision = rev
	return store.get(tiddler)


def _resolve_identifier(id):
	"""
	resolve tiddler identifier string

	the identifier is a string of the form
	<bags|recipes>/<name>/<title>[/<revision>]

	returns type (bag or recipe), name, tiddler title and revision number
	"""
	try:
		type, name, title, rev = id.split("/")
		rev = int(rev)
	except ValueError: # revision implicit
		type, name, title = id.split("/")
		rev = 0 # HEAD
	type = type[:-1] # strip plural
	return type, name, title, rev


def _generate_response(content, environ, start_response):
	serialize_type, mime_type = web.get_serialize_type(environ) # XXX: not suitable here!?
	cache_header = ("Cache-Control", "no-cache") # ensure accessing latest HEAD revision
	content_header = ("Content-Type", mime_type) # XXX: should be determined by diff format
	response = [cache_header, content_header]
	start_response("200 OK", response)
	return [content]


def _html_transform(str):
	"""
	escape a string to be HTML-safe

	also replaces line breaks with BR elements in order to make them visible
	"""
	return cgi.escape(str).replace("\n", " <br />\n")



from setuptools import setup, find_packages

from tiddlywebwiki import __version__ as VERSION


setup(name = 'tiddlywebwiki',
        version = VERSION,
        description = 'A TiddlyWeb plugin to provide a multi-user TiddlyWiki environment.',
        author = 'FND',
        author_email = 'FNDo@gmx.net',
        packages = find_packages(exclude=['test']),
        py_modules = ['differ'],
        scripts = ['twinstance'],
        platforms = 'Posix; MacOS X; Windows',
        install_requires = ['tiddlyweb>=0.9.79',
            'tiddlywebplugins.wikklytextrender',
            'tiddlywebplugins.status',
            'tiddlywebplugins.atom',
            'tiddlywebplugins.utils',
            'BeautifulSoup',
            'wikklytext'],
        include_package_data = True,
        )



"""
Configuration additions for tiddlywebwiki.

base_tiddlywiki -- the file location of the tiddlywiki
file into which Tiddlers are pushed when creating
outgoing TiddlyWiki representations from TiddlyWeb. This
can be an absolute path or relative to the startup
directory of the server.
"""

import os


CLIENT_PLUGIN_NAMES = [
    'TiddlyWebAdaptor.js',
    'ServerSideSavingPlugin.js',
    'TiddlyWebConfig.js',
    'RevisionsCommandPlugin.js',
    'DiffFormatterPlugin.js']


# determine file paths
try:
    from pkg_resources import resource_filename
    BASE_TIDDLYWIKI = resource_filename('tiddlywebwiki', 'empty.html')
    CLIENT_PLUGINS = ['file:%s' % resource_filename('tiddlywebwiki', plugin)
            for plugin in CLIENT_PLUGIN_NAMES]
    # The following list comprehension is required to make sure the meta
    # files get unpacked into the egg cache.
    ['file:%s' % resource_filename('tiddlywebwiki', '%s.meta' % plugin)
            for plugin in CLIENT_PLUGIN_NAMES]
except ImportError:
    BASE_TIDDLYWIKI = os.path.join('tiddlywebwiki', 'empty.html')
    CLIENT_PLUGINS = ['file:%s' % os.path.join('tiddlywebwiki', plugin)
        for plugin in CLIENT_PLUGIN_NAMES]


config = {
        'base_tiddlywiki': BASE_TIDDLYWIKI,
        'instance_tiddlers': {
            'system': CLIENT_PLUGINS
            },
        'extension_types': {
            'wiki': 'text/x-tiddlywiki',
            },
        'serializers': {
            'text/x-tiddlywiki': ['tiddlywebwiki.serialization', 'text/html; charset=UTF-8'],
            },
        'wikitext.default_renderer': 'tiddlywebplugins.wikklytextrender',
        # XXX the following is, in most cases, redundant
        'wikitext.type_render_map': {
            'text/x-tiddlywiki': 'tiddlywebplugins.wikklytextrender',
            }
        }
